/***************************************************************************
* Copyright (c) 2015-2018 Pier Luigi Fiorini <pierluigi.fiorini@gmail.com>
* Copyright (c) 2013 Abdurrahman AVCI <abdurrahmanavci@gmail.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the
* Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
***************************************************************************/

#include "GreeterApp.h"
#include "GreeterProxy.h"
#include "Constants.h"
#include "ScreenModel.h"
#include "UserModel.h"
#include "KeyboardModel.h"
#include "MessageHandler.h"

#include <QCommandLineParser>
#include <QGuiApplication>
#include <QQuickItem>
#include <QQuickView>
#include <QQmlContext>
#include <QQmlEngine>
#include <QDebug>
#include <QTimer>
#include <QTranslator>
#include <sys/types.h>
#include <signal.h>
#include <statgrab.h>
#include "Utils.h"
#include "xservergrabber.h"
#include <xcb/xcb.h>
#include <QX11Info>
#include <X11/Xlib.h>
#include <sys/prctl.h>
#include <QDBusInterface>
#include <QtWidgets/QApplication>
#undef Status
static const QEvent::Type StartupEventType = static_cast<QEvent::Type>(QEvent::registerEventType());

namespace NSS {
    GreeterApp::GreeterApp(QObject *parent)
        : QObject(parent)
    {
        m_components_tranlator = new QTranslator();
        if (m_components_tranlator->load(QLocale::system(), QString(), QString(), QStringLiteral(COMPONENTS_TRANSLATION_DIR)))
            QCoreApplication::installTranslator(m_components_tranlator);

        m_userModel = new UserModel();
        m_keyboard = new KeyboardModel();
    }


    QString GreeterApp::socketName() const
    {
        return m_socket;
    }

    void GreeterApp::setSocketName(const QString &name)
    {
        m_socket = name;
    }
    void GreeterApp::customEvent(QEvent *event)
    {
        if (event->type() == StartupEventType)
            startup();
    }
    bool GreeterApp::nativeEventFilter(const QByteArray &eventType, void *message, long *result)
    {
        if (qstrcmp(eventType, "xcb_generic_event_t") != 0) {
            return false;
        }
        xcb_generic_event_t *event = reinterpret_cast<xcb_generic_event_t*>(message);
        const uint8_t responseType = event->response_type & ~0x80;
        if (responseType == XCB_CONFIGURE_NOTIFY) {
            xcb_configure_notify_event_t *xc = reinterpret_cast<xcb_configure_notify_event_t*>(event);
            for (QQuickView *view : m_views) {
                if(view->winId() == xc->window)
                {
                    return false;
                }
                activatePrimary();
            }

        }else if(responseType == XCB_MAP_NOTIFY){
            xcb_map_notify_event_t *xm = reinterpret_cast<xcb_map_notify_event_t*>(event);
            for (QQuickView *view : m_views) {
                if(view->winId() == xm->window)
                {
                    return false;
                }
                activatePrimary();
            }
        }
        return false;
    }
    void GreeterApp::addViewForScreen(QScreen *screen) {
        // create view
        QQuickView *view = new QQuickView();
        view->setScreen(screen);
        view->setFlags(Qt::X11BypassWindowManagerHint |
                       Qt::WindowStaysOnTopHint |
                       Qt::FramelessWindowHint);
        view->setResizeMode(QQuickView::SizeRootObjectToView);
        view->setGeometry(screen->geometry());

        m_views.append(view);

        connect(qGuiApp, &QGuiApplication::screenRemoved, view, [view, this, screen](QScreen *s) {
            if (s == screen)
                removeViewForScreen(view);
        });

        // always resize when the screen geometry changes
        connect(screen, &QScreen::geometryChanged, this, [view](const QRect &r) {
            view->setGeometry(r);
        });

        view->engine()->addImportPath(QStringLiteral(IMPORTS_INSTALL_DIR));

        // connect proxy signals
        connect(m_proxy, &GreeterProxy::loginSucceeded, view, &QQuickView::close);
        ScreenModel *screenModel = new ScreenModel(screen, view);
        view->rootContext()->setContextProperty(QStringLiteral("screenModel"), screenModel);
        view->rootContext()->setContextProperty(QStringLiteral("userModel"), m_userModel);
        view->rootContext()->setContextProperty(QStringLiteral("nss"), m_proxy);
        view->rootContext()->setContextProperty(QStringLiteral("keyboard"), m_keyboard);
        view->rootContext()->setContextProperty(QStringLiteral("primaryScreen"), QGuiApplication::primaryScreen() == screen);
        view->rootContext()->setContextProperty(QStringLiteral("__nss_errors"), QString());
        view->rootContext()->setContextProperty(QStringLiteral("winId"), static_cast<quint64>(view->winId()));

        // get theme main script
        QUrl mainScriptUrl = QUrl::fromLocalFile(":/theme/Main.qml");
        // set main script as source
        qInfo("Loading %s...", qPrintable(mainScriptUrl.toString()));
        view->setSource(mainScriptUrl);

        // set default cursor
        QCursor cursor(Qt::ArrowCursor);
        view->rootObject()->setCursor(cursor);

        // show
        qDebug() << "Adding view for" << screen->name() << screen->geometry();
        view->showFullScreen();
    }

    void GreeterApp::removeViewForScreen(QQuickView *view) {
        // screen is gone, remove the window
        m_views.removeOne(view);
        view->deleteLater();
    }

    void GreeterApp::startup()
    {
        // Connect to the daemon
        m_proxy = new GreeterProxy(m_userModel->currentUser(), m_socket);
        if (!m_testing && !m_proxy->isConnected()) {
            qCritical() << "Cannot connect to the daemon - is it running?";
            QCoreApplication::exit(EXIT_FAILURE);
            return;
        }
        m_keyboard->setNumLockState(true);
        QList<QScreen *> screens = qGuiApp->primaryScreen()->virtualSiblings();
        for (QScreen *screen : screens)
            addViewForScreen(screen);
        activatePrimary();
        // Handle screens
        connect(qGuiApp, &QGuiApplication::screenAdded, this, &GreeterApp::addViewForScreen);
        connect(qGuiApp, &QGuiApplication::primaryScreenChanged, this, [this](QScreen *) {
            activatePrimary();
        });
        connect(qGuiApp, &QGuiApplication::lastWindowClosed, this, [=]{
            closeGrab();
        });
        establishGrab();
        qApp->installNativeEventFilter(this);
    }
    void GreeterApp::activatePrimary() {
        // activate and give focus to the window assigned to the primary screen
        Q_FOREACH (QQuickView *view, m_views) {
            if (view->screen() == QGuiApplication::primaryScreen()) {
                view->requestActivate();
                XSetInputFocus(QX11Info::display(), view->winId(), RevertToParent, CurrentTime);
                break;
            }
        }
    }
    StartupEvent::StartupEvent()
        : QEvent(StartupEventType)
    {
    }
}
static void sig_usr(int signo)
{
    if (signo == SIGHUP)
    {
        qDebug() << "receive signal SIGHUP";
        qGuiApp->quit();
    }
}
static void startCompositer()
{
    QDBusInterface * compositer = new QDBusInterface(QString::fromLatin1("org.nde.session"),
                                                     QString::fromLatin1("/NdeSession"),
                                                     QString::fromLatin1("org.nde.session"),
                                                     QDBusConnection::sessionBus());
    if(compositer->isValid())
    {
        QDBusMessage reply = compositer->call("getCompositorOpen");
        if(!reply.arguments().isEmpty() && !reply.arguments().first().toBool())
        {
            qputenv("QMLSCENE_DEVICE", "softwarecontext");
        }
    }
    delete compositer;
}
int main(int argc, char **argv)
{
    if(signal(SIGHUP, sig_usr) == SIG_ERR)
        qWarning() << "can't catch SIGHUP";
    prctl(PR_SET_PDEATHSIG, SIGHUP);
    qputenv("QT_IM_MODULE", QByteArray("qtvirtualkeyboard"));
    startCompositer();
    qInstallMessageHandler(NSS::GreeterMessageHandler);
    QGuiApplication::setAttribute(Qt::AA_EnableHighDpiScaling);

    QApplication app(argc, argv);
    QCommandLineParser parser;
    parser.setApplicationDescription(TR("nde-screensaver greeter"));
    parser.addHelpOption();
    parser.addVersionOption();

    QCommandLineOption socketOption(QLatin1String("socket"), TR("Socket name"), TR("name"));
    parser.addOption(socketOption);
    parser.process(app);

    NSS::GreeterApp *greeter = new NSS::GreeterApp();
    greeter->setSocketName(parser.value(socketOption));
    QCoreApplication::postEvent(greeter, new NSS::StartupEvent());
    return app.exec();
}
