#include "ScreenSaverApp.h"
#include "SocketServer.h"

#include "Constants.h"
#include "PowerManager.h"
#include "Utils.h"
#include "lockscreenshortcut.h"
#include "lockscreenmanager.h"
#include <QDebug>
#include <QLocalSocket>
#include "idlewatcher.h"
#include <QProcess>
#include "MessageHandler.h"
#include "screensaver.h"
namespace NSS {

ScreenSaverApp *ScreenSaverApp::self = nullptr;
ScreenSaverApp::ScreenSaverApp(int &argc, char **argv) :
    QApplication(argc, argv),
    m_auth(new Auth (this)),
    m_socketServer(new SocketServer (this)),
    mLockGreeter(new QProcess (this)),
    mLockScreenManager(new LockScreenManager (this)),
    mScreenSaver(new ScreenSaver (this) )
{
    self = this;
    qInstallMessageHandler(NSS::DaemonMessageHandler);
    QSettings setting("nde", "screensaver");
    mIdleTime = setting.value("idleTime", 10 * 60).toInt();
    m_auth->setVerbose(true);
    connect(m_auth, SIGNAL(requestChanged()), this, SLOT(slotRequestChanged()));
    connect(m_auth, SIGNAL(authentication(QString,bool)), this, SLOT(slotAuthenticationFinished(QString,bool)));
    connect(m_auth, SIGNAL(finished(Auth::HelperExitStatus)), this, SLOT(slotHelperFinished(Auth::HelperExitStatus)));
    connect(m_auth, SIGNAL(info(QString,Auth::Info)), this, SLOT(slotAuthInfo(QString,Auth::Info)));
    connect(m_auth, SIGNAL(error(QString,Auth::Error)), this, SLOT(slotAuthError(QString,Auth::Error)));
    connect(this, SIGNAL(loginFailed(QLocalSocket*)), m_socketServer, SLOT(loginFailed(QLocalSocket*)));
    connect(this, SIGNAL(loginSucceeded(QLocalSocket*)), m_socketServer, SLOT(loginSucceeded(QLocalSocket*)));
    connect(this, SIGNAL(loggingIn(QLocalSocket*)), m_socketServer, SLOT(loggingIn(QLocalSocket*)));
    connect(m_socketServer, &SocketServer::login, this, &ScreenSaverApp::login);
    m_socketServer->start(QString(qgetenv("DISPLAY")));
    mShortcut = new LockScreenShortcut;
    connect(mShortcut, &LockScreenShortcut::globalKeyActivated, this, &ScreenSaverApp::lockSession);
    mIdleWatcher = new IdleWatcher(mIdleTime, this);
    mPowerManager = new PowerManager (this);
    connect(mIdleWatcher, &IdleWatcher::requestLock, this, &ScreenSaverApp::lockSession);
    connect(mLockGreeter, QOverload<int, QProcess::ExitStatus>::of(&QProcess::finished),
            this, [=](int exitCode, QProcess::ExitStatus exitStatus){
        Q_UNUSED(exitCode)
        if(exitStatus == QProcess::CrashExit)
        {
            mIdleWatcher->reset(mIdleTime);
        }
    });
    if (mLockScreenManager->startup(true, 0))
        qDebug() << "LockScreenManager started successfully";
    else
        qWarning() << "LockScreenManager couldn't start";
    unLock();
}
ScreenSaverApp::~ScreenSaverApp()
{
    m_socketServer->stop();
    mShortcut->deleteLater();
    mLockGreeter->kill();
}
void ScreenSaverApp::setIdleTime(int secs)
{
    mIdleTime = secs;
    mIdleWatcher->reset(mIdleTime);
}
void ScreenSaverApp::lockSession()
{
    if(mLockGreeter->state() == QProcess::Running ||
            mLockGreeter->state() == QProcess::Starting)
        return;
    mLockGreeter->start(QString("%1/nde-screensaver-greeter").arg(QStringLiteral(LIBEXEC_INSTALL_DIR)),
                        QStringList {} << "--socket" << m_socketServer->socketAddress());
    mLockScreenManager->saveLockStatus(LockScreenManager::LOCKED);
    mIdleWatcher->removeIdleTimeout();
}
void ScreenSaverApp::unLock()
{
    daemonInterface = new com::vm::daemon("com.vm.daemon",
                                    "/daemon",QDBusConnection::systemBus(), this);
    mLockScreenManager->saveLockStatus(LockScreenManager::UNLOCKED);
    //token解锁
    connect(daemonInterface, &com::vm::daemon::tokenSuccess, this, [=]{
        qDebug() << "ScreenSaverApp: unlock Screen";
        mLockGreeter->kill();
        mLockScreenManager->saveLockStatus(LockScreenManager::UNLOCKED);
        mIdleWatcher->reset(mIdleTime);
    });
    //密码解锁
    connect(daemonInterface, &com::vm::daemon::autoLogin, this, [=]{
        qWarning() << "ScreenSaverApp: auto login";
        QString userName = daemonInterface->userName();
        QString password = daemonInterface->password();
        m_socket = m_socketServer->currentSocket();
        if(m_socket && locked())
        {
            login(m_socket, userName, password);
        }
    });

}
void ScreenSaverApp::login(QLocalSocket *socket,
                           const QString &user, const QString &password) {
    m_socket = socket;
    emit loggingIn(m_socket);
    m_passPhrase = password;

    m_auth->setUser(user);
    m_auth->start();
}

void ScreenSaverApp::slotAuthenticationFinished(const QString &user, bool success) {
    Q_UNUSED(user)
    if (success) {
        qDebug() << "Authenticated successfully";
        if (m_socket)
        {
            emit loginSucceeded(m_socket);
            mLockScreenManager->saveLockStatus(LockScreenManager::UNLOCKED);
            mIdleWatcher->reset(mIdleTime);
        }
    } else if (m_socket) {
        qDebug() << "Authentication failure";
        emit loginFailed(m_socket);
    }
    m_socket = nullptr;
}

void ScreenSaverApp::slotAuthInfo(const QString &message, Auth::Info info) {
    // TODO: presentable to the user, eventually
    Q_UNUSED(info)
    qWarning() << "Authentication information:" << message;
}

void ScreenSaverApp::slotAuthError(const QString &message, Auth::Error error) {
    // TODO: handle more errors
    qWarning() << "Authentication error:" << message;
    emit loginFailed(m_socket);
}

void ScreenSaverApp::slotHelperFinished(Auth::HelperExitStatus status) {
    Q_UNUSED(status)
}

void ScreenSaverApp::slotRequestChanged() {
    if (m_auth->request()->prompts().length() == 1) {
        m_auth->request()->prompts()[0]->setResponse(qPrintable(m_passPhrase));
        m_auth->request()->done();
    } else if (m_auth->request()->prompts().length() == 2) {
        m_auth->request()->prompts()[0]->setResponse(qPrintable(m_auth->user()));
        m_auth->request()->prompts()[1]->setResponse(qPrintable(m_passPhrase));
        m_auth->request()->done();
    }
}
PowerManager *ScreenSaverApp::powerManager() const {
    return mPowerManager;
}
bool ScreenSaverApp::locked() const
{
    qWarning() << mLockGreeter->state();
    if(mLockGreeter->state() == QProcess::NotRunning)
        return false;
    if(mLockGreeter->state() == QProcess::Running ||
            mLockGreeter->state() == QProcess::Starting)
        return true;
    return false;
}
}
int main(int argc, char **argv) {
    NSS::ScreenSaverApp app(argc, argv);
    return app.exec();
}
