/***************************************************************************
* Copyright (c) 2015 Pier Luigi Fiorini <pierluigi.fiorini@gmail.com>
* Copyright (c) 2013 Abdurrahman AVCI <abdurrahmanavci@gmail.com>
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the
* Free Software Foundation, Inc.,
* 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
***************************************************************************/

#include "SocketServer.h"
#include "ScreenSaverApp.h"
#include "Messages.h"
#include "PowerManager.h"
#include "SocketWriter.h"
#include "Utils.h"

#include <QLocalServer>

namespace NSS {
    SocketServer::SocketServer(QObject *parent) : QObject(parent) {
    }

    QString SocketServer::socketAddress() const {
        if (m_server)
            return m_server->fullServerName();
        return QString();
    }
    QLocalSocket * SocketServer::currentSocket() const
    {
        if(m_socket)
            return m_socket;
        return nullptr;
    }
    bool SocketServer::start(const QString &displayName) {
        // check if the server has been created already
        if (m_server)
            return false;

        QString socketName = QStringLiteral("nss-%1-%2").arg(displayName).arg(generateName(6));

        // log message
        qDebug() << "Socket server starting..." << socketName;

        // create server
        m_server = new QLocalServer(this);

        // start listening
        if (!m_server->listen(socketName)) {
            // log message
            qCritical() << "Failed to start socket server.";

            // return fail
            return false;
        }


        // log message
        qDebug() << "Socket server started.";

        // connect signals
        connect(m_server, SIGNAL(newConnection()), this, SLOT(newConnection()));

        // return success
        return true;
    }

    void SocketServer::stop() {
        // check flag
        if (!m_server)
            return;

        // log message
        qDebug() << "Socket server stopping...";

        // delete server
        m_server->deleteLater();
        m_server = nullptr;

        // log message
        qDebug() << "Socket server stopped.";
    }

    void SocketServer::newConnection() {
        // get pending connection
        m_socket = m_server->nextPendingConnection();
        qDebug() << "new connection coming";
        // connect signals
        connect(m_socket, SIGNAL(readyRead()), this, SLOT(readyRead()));
        connect(m_socket, SIGNAL(disconnected()), m_socket, SLOT(deleteLater()));
    }

    void SocketServer::readyRead() {
        QLocalSocket *socket = qobject_cast<QLocalSocket *>(sender());

        if (!socket)
            return;
        QDataStream input(socket);
        quint32 message;
        QString user, password;

        input >> message;
        input >>  user >> password;

        switch (GreeterMessages(message)) {
            case GreeterMessages::Connect: {
                qDebug() << "Message received from greeter: Connect";
                SocketWriter(socket) << quint32(DaemonMessages::Capabilities) << quint32(SaverApp->powerManager()->capabilities());
                emit connected();
            }
            break;
            case GreeterMessages::Login: {
                qDebug() << "Message received from greeter: Login";
                if(!user.isEmpty() && !password.isEmpty())
                    emit login(socket, user, password);
            }
            break;
            case GreeterMessages::PowerOff: {
                qDebug() << "Message received from greeter: PowerOff";
                SaverApp->powerManager()->powerOff();
            }
            break;
            case GreeterMessages::Reboot: {
                // log message
                qDebug() << "Message received from greeter: Reboot";

                // reboot
                SaverApp->powerManager()->reboot();
            }
            break;
            case GreeterMessages::Suspend: {
                // log message
                qDebug() << "Message received from greeter: Suspend";

                // suspend
                SaverApp->powerManager()->suspend();
            }
            break;
            case GreeterMessages::Hibernate: {
                // log message
                qDebug() << "Message received from greeter: Hibernate";

                // hibernate
                SaverApp->powerManager()->hibernate();
            }
            break;
            case GreeterMessages::HybridSleep: {
                // log message
                qDebug() << "Message received from greeter: HybridSleep";

                // hybrid sleep
                SaverApp->powerManager()->hybridSleep();
            }
            break;
            default: {
                qWarning() << "Unknown message" << message;
            }
        }
    }

    void SocketServer::loginFailed(QLocalSocket *socket) {
        SocketWriter(socket) << quint32(DaemonMessages::LoginFailed);
    }

    void SocketServer::loginSucceeded(QLocalSocket *socket) {
        SocketWriter(socket) << quint32(DaemonMessages::LoginSucceeded);
    }
    void SocketServer::loggingIn(QLocalSocket *socket) {
        SocketWriter(socket) << quint32(DaemonMessages::LoggingIn);
    }
}
