#include "lockscreenmanager.h"
#include <QDBusMetaType>
#include <QTimer>
#include <QDBusReply>
#include <QDebug>
#include <QDBusUnixFileDescriptor>
#include <unistd.h>
#include <QFile>
#include <Nde/Globals>
#include "ScreenSaverApp.h"

namespace NSS {
LockScreenManager::LockScreenManager(QObject *parent) :
    QObject(parent)
    , mProvider{nullptr}
    , mLockedBeforeSleep{false}
{
}

LockScreenManager::~LockScreenManager()
{
    delete mProvider;
}

bool LockScreenManager::startup(bool lockBeforeSleep, int powerAfterLockDelay)
{
    Q_UNUSED(powerAfterLockDelay)
    mProvider = new LogindProvider;
    if (!mProvider->isValid())
    {
        mProvider->deleteLater();
        mProvider = new ConsoleKit2Provider;
        if (!mProvider->isValid())
        {
            mProvider->deleteLater();
            mProvider = nullptr;
        }
    }
    saveLockStatus(UNLOCKED);
    if (!mProvider)
    {
        qDebug() << "LockScreenManager: no valid provider";
        return false;
    }
    qDebug() << "LockScreenManager:"
            << mProvider->metaObject()->className()
            << "will be used";

    connect(mProvider, &LockScreenProvider::lockRequested, [] {
        qDebug() << "LockScreenManager: lock requested";
        SaverApp->lockSession();
    });

    if (lockBeforeSleep)
    {
        connect(mProvider,
                &LockScreenProvider::aboutToSleep,
                [this] (bool beforeSleep)
        {
            if (beforeSleep)
            {
                qDebug() << "LockScreenManager: system is about to sleep";

                mLockedBeforeSleep = true;
                SaverApp->lockSession();
                qDebug() << "LockScreenManager: after lockScreen";
            } else
            {
                inhibit();
            }
        });

        inhibit();
    }

    return true;
}
void LockScreenManager::saveLockStatus(LOCK_STATUS status)
{
    QString lockStatus = "/etc/lock-status.conf";
    QFile file(lockStatus);
    if(file.open(QIODevice::WriteOnly | QIODevice::Text))
    {
        QTextStream out(&file);
        out << "status=" << status;
    }
    file.close();
}
void LockScreenManager::inhibit()
{
    if (!mProvider->inhibit())
        qDebug() << "LockScreenManager: could not inhibit session provider";
}

/*
 * Logind Provider
 */

LogindProvider::LogindProvider() :
    mInterface(QStringLiteral("org.freedesktop.login1"),
            QStringLiteral("/org/freedesktop/login1"),
            QStringLiteral("org.freedesktop.login1.Manager"),
            QDBusConnection::systemBus())
{
    if (mInterface.isValid())
    {
        connect(&mInterface, SIGNAL(PrepareForSleep(bool)),
                this, SIGNAL(aboutToSleep(bool)));

        QString sessionId = QDBusInterface(
                QStringLiteral("org.freedesktop.login1"),
                QStringLiteral("/org/freedesktop/login1/session/self"),
                QStringLiteral("org.freedesktop.login1.Session"),
                QDBusConnection::systemBus())
            .property("Id").toString();

        // listen to Lock signal as well
        mInterface.connection().connect(
                QStringLiteral("org.freedesktop.login1"),
                QStringLiteral("/org/freedesktop/login1/session/") + sessionId,
                QStringLiteral("org.freedesktop.login1.Session"),
                QStringLiteral("Lock"),
                this,
                SIGNAL(lockRequested()));
    }
}

LogindProvider::~LogindProvider()
{
    release();
}

bool LogindProvider::isValid()
{
    return mInterface.isValid();
}

bool LogindProvider::inhibit()
{
    if (mFileDescriptor)
        return false;

    QDBusReply<QDBusUnixFileDescriptor> reply = mInterface.call(
            QStringLiteral("Inhibit"),
            QStringLiteral("sleep"),
            QStringLiteral("Nde Session"),
            QStringLiteral("Start screen locker before sleep."),
            QStringLiteral("delay"));

    if (!reply.isValid())
    {
        qDebug() << "LockScreenManager: " << reply.error();
        return false;
    }

    mFileDescriptor.reset(new QDBusUnixFileDescriptor{reply.value()});
    return true;
}

void LogindProvider::release()
{
    mFileDescriptor.reset(nullptr);
}

/*
 * ConsoleKit2 provider
 */

ConsoleKit2Provider::ConsoleKit2Provider() :
    mInterface(QStringLiteral("org.freedesktop.ConsoleKit"),
            QStringLiteral("/org/freedesktop/ConsoleKit/Manager"),
            QStringLiteral("org.freedesktop.ConsoleKit.Manager"),
            QDBusConnection::systemBus()),
    mMethodInhibitPresent(false)

{
    // validate interface, look for melhod Inhibit
    if (mInterface.isValid())
    {
        // this is needed to differentiate ConsoleKit2 from ConsoleKit
        QDBusReply<QString> reply = QDBusInterface(
                QStringLiteral("org.freedesktop.ConsoleKit"),
                QStringLiteral("/org/freedesktop/ConsoleKit/Manager"),
                QStringLiteral("org.freedesktop.DBus.Introspectable"),
                QDBusConnection::systemBus())
            .call(QStringLiteral("Introspect"));
        mMethodInhibitPresent = reply.value().contains(QL1S("Inhibit"));

        if (mMethodInhibitPresent)
        {
            connect(&mInterface, SIGNAL(PrepareForSleep(bool)),
                    this, SIGNAL(aboutToSleep(bool)));

            QDBusReply<QDBusObjectPath> sessionObjectPath = mInterface
                .call(QSL("GetCurrentSession"));

            // listen to Lock signal as well
            mInterface.connection().connect(
                    QStringLiteral("org.freedesktop.ConsoleKit"),
                    sessionObjectPath.value().path(),
                    QStringLiteral("org.freedesktop.ConsoleKit.Session"),
                    QStringLiteral("Lock"),
                    this,
                    SIGNAL(lockRequested()));
        }
    }
}

ConsoleKit2Provider::~ConsoleKit2Provider()
{
    release();
}

bool ConsoleKit2Provider::isValid()
{
    return mInterface.isValid() && mMethodInhibitPresent;
}

bool ConsoleKit2Provider::inhibit()
{
    if (mFileDescriptor)
        return false;

    QDBusReply<QDBusUnixFileDescriptor> reply = mInterface.call(
            QStringLiteral("Inhibit"),
            QStringLiteral("sleep"),
            QStringLiteral("Nde Power Management"),
            QStringLiteral("Start screen locker before sleep."),
            QStringLiteral("delay"));

    if (!reply.isValid())
    {
        qDebug() << "LockScreenWatcher: " << reply.error();
        return false;
    }

    mFileDescriptor.reset(new QDBusUnixFileDescriptor{reply.value()});
    return true;
}

void ConsoleKit2Provider::release()
{
    mFileDescriptor.reset(nullptr);
}
}
